#ifndef _compiler_H_
#define _compiler_H_

#include <iostream>
#include <string>
#include <vector>
#include <map>
#include <stdexcept>
#include <set>

class syntax_error : public std::runtime_error {
public:
    syntax_error(const std::string& what_arg, size_t lineno)
        : std::runtime_error {what_arg}, _lineno {lineno} {}
    size_t lineno() const { return _lineno; }
private:
    size_t _lineno {0};
};

class compiler;

using operation = void(*)(compiler&, const std::string&);
struct function_call {
    operation func;
    std::string arg;

    function_call(operation, const std::string&);
};

struct ref {
    std::string name;
    size_t prog_index;
    size_t lineno;

    ref(const std::string& n, size_t i, size_t l);
};

class compiler {
public:
    compiler();

    std::string eval(std::istream& in, bool& done);

private:
    std::map<std::string, operation> _ops; // operation registry
    std::vector<function_call> _prog; // compiled program
    std::map<std::string, size_t> _labels; // label name => _prog index
    std::vector<ref> _label_refs; // used during compilation to resolve label references
    std::vector<std::string> _values; // value stack
    std::vector<size_t> _calls; // call stack
    std::map<std::string, std::string> _vars; // variable name => string value
    std::vector<ref> _var_refs; // used during compilation to resolve variable references
    size_t _pc {0}; // current program counter (index into _prog)
    bool _done {false};

    void reset();

    int pop_int();
    std::string pop_str();

    // stack operations
    static void push_literal(compiler&, const std::string& val);
    static void assign(compiler&, const std::string& varname);
    static void push_variable(compiler&, const std::string& varname);

    // integer operations
    static void add(compiler&, const std::string&);
    static void sub(compiler&, const std::string&);
    static void mul(compiler&, const std::string&);
    static void div(compiler&, const std::string&);
    static void mod(compiler&, const std::string&);
    static void abs(compiler&, const std::string&);
    static void neg(compiler&, const std::string&);
    static void inc(compiler&, const std::string&);
    static void dec(compiler&, const std::string&);

    // string operations
    static void dup(compiler&, const std::string&);
    static void rev(compiler&, const std::string&);
    static void slc(compiler&, const std::string&);
    static void idx(compiler&, const std::string&);
    static void cat(compiler&, const std::string&);
    static void len(compiler&, const std::string&);
    static void rot(compiler&, const std::string&);
    static void enl(compiler&, const std::string&);

    // tests & jumps
    static void gto(compiler&, const std::string&);
    static void geq(compiler&, const std::string&);
    static void gne(compiler&, const std::string&);
    static void glt(compiler&, const std::string&);
    static void gle(compiler&, const std::string&);
    static void ggt(compiler&, const std::string&);
    static void gge(compiler&, const std::string&);

    // subroutines
    static void fun(compiler&, const std::string&);
    static void ret(compiler&, const std::string&);

    // I/O
    static void inp(compiler&, const std::string&);
    static void out(compiler&, const std::string&);
    static void err(compiler&, const std::string&);

    // exit
    static void end(compiler&, const std::string&);
};

#endif // _compiler_H_
